#!/usr/bin/python
########################################################################
#
# File:   fbcore.py
# Date:   24.05.2005
#
# Contents:
#   Basic Firebird classes.
#
# Contributors: Pavel Cisar
#
# Permission is hereby granted, free of charge, to any person
# obtaining a copy of this software and associated documentation files
# (the "Software"), to deal in the Software without restriction,
# including without limitation the rights to use, copy, modify, merge,
# publish, distribute, sublicense, and/or sell copies of the Software,
# and to permit persons to whom the Software is furnished to do so,
# subject to the following conditions:
#
# The above copyright notice and this permission notice shall be
# included in all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
# EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
# MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
# NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS
# BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN
# ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
# conECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
########################################################################

########################################################################
# imports
########################################################################

import sys
import kinterbasdb as kdb
import unittest
from ibptoolbox import LateBindingProperty

########################################################################
# constants
########################################################################

FBCORE_VERSION = "0.2"

########################################################################
# classes
########################################################################

class FBMDatabase(object):
    """This class represents a database.

       in contains methods to attach, create or drop database, properties
       and methods to access metadata or create new metadata objects.
    """

    dsn = None
    user = None

    def __init__(self,con=None):
        """Constructor.
           con is kinterbasdb.Connection object. Optional.
        """
        super(FBMDatabase,self).__init__(self)
        self.__domains=None
        self.__tables=None
        self.__views=None
        self.__generators=None
        self.__character_sets=None
        self.__collations=None
        self.__exceptions=None
        self.__indices=None
        self.__constraints=None
        self._constraintIndices=None
        self.__triggers=None
        if con:
            self.con = con
            self.__c = self.con.cursor()

    #--- Basic Database manipulation routines

    def commit(self):
        """Commit transaction.
        """
        self.con.commit()
    def rollback(self):
        """Rollback transaction.
        """
        self.con.rollback()
    def createDatabase(self,cmd,dialect=3):
        """Creates new database.
           cmd equals to cmd parameter to kinterbasdb.create_database cmd parameter.
        """
        self.con = kdb.create_database(cmd,dialect)
        self.__c = self.con.cursor()
        self.reload()
    def attachDatabase(self,**args):
        """Attach to existing database.
           **args is a dictionary of arguments for kinterbasdb.connect function.
        """
        self.con = kdb.connect(**args)
        self.__c = self.con.cursor()
        self.reload()
        self.dsn = args.get('dsn')
        self.user = args.get('user')
    def close(self):
        """Close connection to the database.
        """
        if self.con:
            self.con.close()
            self.con = None
    def drop(self):
        """Drop attached database.
        """
        if self.con:
            self.con.drop_database()
            self.con = None
    def cursor(self):
        return self.con.cursor()
    def reload(self,data=None):
        """Forces reload of database metadata.
           data - if not defined or None (default) forces reload of all metadata.
           Could contain name of one (and only one) from metadata lists:
               tables
               views
               generators
               characterSets
               collations
               exceptions
               domains
           reload is not immediate, but appears at next reference to
           "reloaded" property.
        """
        if (data == "tables" or not data):
            self.__tables = None
        if (data == "generators" or not data):
            self.__generators = None
        if (data == "characterSets" or not data):
            self.__character_sets = None
        if (data == "collations" or not data):
            self.__collations = None
        if (data == "exceptions" or not data):
            self.__exceptions = None
        if (data == "domains" or not data):
            self.__domains = None
        if (data == "views" or not data):
            self.__views = None
        if (data == "indices" or not data):
            self.__indices = None
        if (data == "constraints" or not data):
            self.__constraints = None
            self.__constraintIndices = None
        if (data == "triggers" or not data):
            self.__triggers = None

    #--- private
    def __objectByName(self,list,name):
        for o in list:
            if o.name == name:
                return o
        return None

    #--- protected
    def _selectRow(self,cmd,commit=True):
        self.__c.execute(cmd)
        row = dict(self.__c.fetchonemap().items())
        if commit:
            self.con.commit()
        return row

    #--- special attribute access methods

    def _get_alltables(self):
        if self.__tables is None:
            self.__c.execute("select * from rdb$relations where rdb$view_blr is null")
            self.__tables = [FBMTable(self,**dict(row.items())) for row in self.__c.itermap()]
            self.con.commit()
        return [o for o in self.__tables if not o.dropped]
    def _get_tables(self):
        return [t for t in self._get_alltables() if not t.isSystemObj]
    def _get_systables(self):
        return [t for t in self._get_alltables() if t.isSystemObj]
    def _get_views(self):
        if self.__views is None:
            self.__c.execute("select * from rdb$relations where rdb$view_blr is not null")
            self.__views = [FBMView(self,**dict(row.items())) for row in self.__c.itermap()]
            self.con.commit()
        return [o for o in self.__views if not o.dropped]
    def _get_allgenerators(self):
        if self.__generators is None:
            self.__c.execute("select * from rdb$generators")
            self.__generators = [FBMGenerator(self,**dict(row.items())) for row in self.__c.itermap()]
            self.con.commit()
        return [o for o in self.__generators if not o.dropped]
    def _get_generators(self):
        return [g for g in self._get_allgenerators() if not g.isSystemObj]
    def _get_sysgenerators(self):
        return [g for g in self._get_allgenerators() if g.isSystemObj]
    def _get_characterSets(self):
        if self.__character_sets is None:
            self.__c.execute("select * from rdb$character_sets")
            self.__character_sets = [FBMCharacterSet(self,**dict(row.items())) for row in self.__c.itermap()]
            self.con.commit()
        return self.__character_sets
    def _get_collations(self):
        if self.__collations is None:
            self.__c.execute("select * from rdb$collations")
            self.__collations = [FBMCollation(self,**dict(row.items())) for row in self.__c.itermap()]
            self.con.commit()
        return self.__collations
    def _get_exceptions(self):
        if self.__exceptions is None:
            self.__c.execute("select * from rdb$exceptions")
            self.__exceptions = [FBMException(self,**dict(row.items())) for row in self.__c.itermap()]
            self.con.commit()
        return [o for o in self.__exceptions if not o.dropped]
    def _get_alldomains(self):
        if self.__domains is None:
            self.__c.execute("select * from rdb$fields")
            self.__domains = [FBMDomain(self,**dict(row.items())) for row in self.__c.itermap()]
            self.con.commit()
        return [o for o in self.__domains if not o.dropped]
    def _get_domains(self):
        return [d for d in self._get_alldomains() if not d.isSystemObj]
    def _get_sysdomains(self):
        return [d for d in self._get_alldomains() if d.isSystemObj]
    def _get_allindices(self):
        if self.__indices is None:
            self.__c.execute("select * from rdb$indices")
            self.__indices = [FBMIndex(self,**dict(row.items())) for row in self.__c.itermap()]
            self.con.commit()
        return [o for o in self.__indices if not o.dropped]
    def _get_indices(self):
        return [i for i in self._get_allindices() if not i.isSystemObj]
    def _get_sysindices(self):
        return [i for i in self._get_allindices() if i.isSystemObj]
    def _get_constraints(self):
        if self.__constraints is None:
            self.__c.execute("select * from rdb$relation_constraints")
            self.__constraints = [FBMConstraint(self,**dict(row.items())) for row in self.__c.itermap()]
            self.con.commit()
            self.__constraintIndices = {}
            for c in self.__constraints:
                if c.index_name is not None:
                    self.__constraintIndices[c.index_name.strip()] = c.relation_name.strip()
        return [o for o in self.__constraints if not o.dropped]
    def _get_constraintIndices(self):
        if self.__constraintIndices is None:
            self.__constraintIndices = {}
            for c in self.constraints:
                if c.index_name is not None:
                    self.__constraintIndices[c.index_name.strip()] = c.relation_name.strip()
        return self.__constraintIndices
    def _get_alltriggers(self):
        if self.__triggers is None:
            self.__c.execute("select * from rdb$triggers")
            self.__triggers = [FBMTrigger(self,**dict(row.items())) for row in self.__c.itermap()]
            self.con.commit()
        return [o for o in self.__triggers if not o.dropped]
    def _get_triggers(self):
        return [t for t in self._get_alltriggers() if not t.isSystemObj]
    def _get_systriggers(self):
        return [t for t in self._get_alltriggers() if t.isSystemObj]

    #--- properties

    tables = LateBindingProperty(_get_tables,None,None,"List of all user tables in database.\nItems are FBMTable objects.")
    systables = LateBindingProperty(_get_systables,None,None,"List of all system tables in database.\nItems are FBMTable objects.")
    views = LateBindingProperty(_get_views,None,None,"List of all views in database.\nItems are FBMView objects.")
    indices = LateBindingProperty(_get_indices,None,None,"List of all user indices in database.\nItems are FBMIndex objects.")
    sysindices = LateBindingProperty(_get_sysindices,None,None,"List of all system indices in database.\nItems are FBMIndex objects.")
    generators = LateBindingProperty(_get_generators,None,None,"List of all user generators in database.\nItems are FBMGenerator objects.")
    sysgenerators = LateBindingProperty(_get_sysgenerators,None,None,"List of all system generators in database.\nItems are FBMGenerator objects.")
    characterSets = LateBindingProperty(_get_characterSets,None,None,"List of all character sets in database.\nItems are FBMCharacterSet objects.")
    collations = LateBindingProperty(_get_collations,None,None,"List of all collations in database.\nItems are FBMCollation objects.")
    exceptions = LateBindingProperty(_get_exceptions,None,None,"List of all exceptions in database.\nItems are FBMException objects.")
    domains = LateBindingProperty(_get_domains,None,None,"List of all user domains in database.\nItems are FBMDomain objects.")
    sysdomains = LateBindingProperty(_get_sysdomains,None,None,"List of all system domains in database.\nItems are FBMDomain objects.")
    constraints = LateBindingProperty(_get_constraints,None,None,"List of all constraints in database.\nItems are FBMConstraint objects.")
    triggers = LateBindingProperty(_get_triggers,None,None,"List of all triggers in database.\nItems are FBMTrigger objects.")

    #--- functions to access metadata by name

    def domain(self,name):
        """Returns handler object for domain of specified name
           or None if it doesn't exists."""
        return self.__objectByName(self.domains,name)
    def table(self,name):
        """Returns handler object for table of specified name
           or None if it doesn't exists."""
        return self.__objectByName(self.tables,name)
    def view(self,name):
        """Returns handler object for view of specified name
           or None if it doesn't exists."""
        return self.__objectByName(self.views,name)
    def generator(self,name):
        """Returns handler object for generator of specified name
           or None if it doesn't exists."""
        return self.__objectByName(self.generators,name)
    def charcterSet(self,name):
        """Returns handler object for character set of specified name
           or None if it doesn't exists."""
        return self.__objectByName(self.characterSets,name)
    def collation(self,name):
        """Returns handler object for collation of specified name
           or None if it doesn't exists."""
        return self.__objectByName(self.collations,name)
    def exception(self,name):
        """Returns handler object for exception of specified name
           or None if it doesn't exists."""
        return self.__objectByName(self.exceptions,name)
    def index(self,name):
        """Returns handler object for index of specified name
           or None if it doesn't exists."""
        return self.__objectByName(self.indices,name)

    #--- functions to create new metadata objects

    def newGenerator(self,name,value=None):
        """Create new generator.

           name     Name for the generator.
           value    Value for the new generator. Optional.
        """
        self.con.execute_immediate("create generator %s" % name)
        self.con.commit()
        g = FBMGenerator(self)
        g.load(name)
        if value:
            g.currentValue = value
        if self.__generators is not None:
            self.__generators.append(g)
        return g
    def newException(self,name,message):
        """Create new exception.

           name     Name for the Exception.
           message  Exception message.
        """
        self.con.execute_immediate("create exception %s '%s'" % (name,message))
        self.con.commit()
        e = FBMException(self)
        e.load(name)
        return e
    def newDomain(self,name,datatype,**args):
        """Create new domain.

           name      Name for the domain.
           datatype  SQL name for domain's datatype (for example CHAR(15), INTEGER).
           **args    dict of additional arguments. recognized ones are:
                     characterSet   Character set name.
                     collate        Collation name.
                     notNull        Value is not important. if defined, the domain
                                    would be not NULL.
                     check          Check constrains.
                     default        SQL expression for default.
        """
        args["name"]=name
        args["datatype"]=datatype
        if args.has_key("characterSet"):
            args["characterSet"]="CHARACTER SET %(characterSet)s" % args
        else:
            args["characterSet"]=""
        if args.has_key("collate"):
            args["collate"]="COLLATE %(collate)s" % args
        else:
            args["collate"]=""
        if args.has_key("notNull"):
            args["notNull"]="NOT NULL"
        else:
            args["notNull"]=""
        if args.has_key("check"):
            args["check"]="CHECK (%(check)s)" % args
        else:
            args["check"]=""
        if args.has_key("default"):
            args["default"]="DEFAULT %(default)s" % args
        else:
            args["default"]=""
        s = "create domain %(name)s as %(datatype)s %(characterSet)s %(default)s %(notNull)s %(check)s %(collate)s" % args
        print s
        self.con.execute_immediate(s)
        self.con.commit()
        d = FBMDomain(self)
        d.load(name)
        return d

class FBMBaseDBObject(object):
    _refreshSQL = None
    _dropSQL = None
    nameAttribute = None
    dropped = None

    def __init__(self,db,**args):
        super(FBMBaseDBObject,self).__init__(**args)
        self.__metadata = args
        self.db = db
        self.con = db.con
        self.dropped = not bool(self.__metadata)
        for attr in [a.split("$")[1].lower() for a in self.__metadata.keys()]:
            # attribute is defined in descendant class!
            setattr (self.__class__,attr,property(self.__getmethod(attr)))

    #--- private
    def __getmethod(self,attrname):
        def _getmethod(self):
            if self.__metadata is None:
                return None
            else:
                return self.__metadata["RDB$"+attrname.upper()]
        d = _getmethod
        d.__dict__['attrname'] = attrname
        return d
    def __selectRow(self,cmd,commit=True):
        return self.db._selectRow(cmd,commit)

    #--- protected
    def _get_name(self):
        if (self.nameAttribute != None) and self.__metadata:
            return self.__metadata["RDB$"+self.nameAttribute.upper()].strip()
        else:
            return None
    def _get_short_description(self):
        return self.name
    def _get_full_description(self):
        return "%s (%s)" % (self.name,self.__class__.__name__)
    def _get_isSystemObj(self):
        return False

    #--- properties

    name = LateBindingProperty(_get_name,None,None,"Returns value of nameAttribute (object ID) or None if it's not defined.")
    shortDescription = LateBindingProperty(_get_short_description,None,None,"Returns short description for the instance.")
    fullDescription = LateBindingProperty(_get_full_description,None,None,"Returns full description for the instance.")
    isSystemObj = LateBindingProperty(_get_isSystemObj,None,None,"True if it's a system object.")

    #--- public
    def reload(self):
        if not self.dropped:
            self.load(self.name)

    def load(self, name):
        assert self._refreshSQL, name
        self.__metadata = self.__selectRow(self._refreshSQL % name)
        self.dropped = not bool(self.__metadata)

    def drop(self):
        if self._dropSQL != None:
            self.con.execute_immediate(self._dropSQL % self.name)
            self.con.commit()
            self.dropped = True
            self.__metadata = None

class FBMTable(FBMBaseDBObject):
    def __init__(self,db,**args):
        super(FBMTable,self).__init__(db,**args)
        self._refreshSQL = "select * from rdb$relations where rdb$relation_name='%s'"
        self._dropSQL = "drop table %s"
        self.nameAttribute = "relation_name"

    #--- protected
    def _get_isSystemObj(self):
        return self.system_flag != 0
    def _get_indices(self):
        return [i for i in self.db.indices if i.relation_name.strip() == self.name]

    #--- properties
    indices = LateBindingProperty(_get_indices,None,None,"Returns list of indices defined for table.\nItems are FBMIndex objects.")

class FBMView(FBMBaseDBObject):
    def __init__(self,db,**args):
        super(FBMView,self).__init__(db,**args)
        self._refreshSQL = "select * from rdb$relations where rdb$relation_name='%s'"
        self._dropSQL = "drop view %s"
        self.nameAttribute = "relation_name"

class FBMDomain(FBMBaseDBObject):
    def __init__(self,db,**args):
        super(FBMDomain,self).__init__(db,**args)
        self._refreshSQL = "select * from rdb$fields where rdb$field_name='%s'"
        self._dropSQL = "drop domain %s"
        self.nameAttribute = "field_name"

    #--- protected
    def _get_isSystemObj(self):
        return (self.system_flag == 1) or self.field_name.startswith('RDB$')
    def _get_nullable(self):
        return not self.null_flag

    #--- properties
    nullable = LateBindingProperty(_get_nullable,None,None,"True if domain is not defined with NOT NULL.")

class FBMGenerator(FBMBaseDBObject):
    def __init__(self,db,**args):
        super(FBMGenerator,self).__init__(db,**args)
        self._refreshSQL = "select * from rdb$generators where rdb$generator_name='%s'"
        self._dropSQL = "drop generator %s"
        self.nameAttribute = "generator_name"

    #--- protected
    def _get_value(self):
        c = self.con.cursor()
        s = "select gen_id(%s,0) from rdb$database" % self.generator_name
        c.execute(s)
        res = c.fetchone()[0]
        self.con.commit()
        return res
    def _set_value(self,value):
        self.con.execute_immediate("set generator %s to %d" % (self.generator_name, value))
        self.con.commit()
    def _get_isSystemObj(self):
        return self.system_flag == 1

    #--- properties
    currentValue = LateBindingProperty(_get_value,_set_value,None,"Current generator value.")

class FBMCharacterSet(FBMBaseDBObject):
    def __init__(self,db,**args):
        super(FBMCharacterSet,self).__init__(db,**args)
        self._refreshSQL = "select * from rdb$character_sets where rdb$character_set_name='%s'"
        self.nameAttribute = "character_set_name"

    #--- protected
    def _get_isSystemObj(self):
        return self.system_flag == 1
    def _get_collations(self):
        r = [c for c in self.db.collations if c.character_set_id == self.character_set_id]
        return r
    #--- properties
    collations = LateBindingProperty(_get_collations,None,None,"List of Collations associated with character set.")

class FBMCollation(FBMBaseDBObject):
    def __init__(self,db,**args):
        super(FBMCollation,self).__init__(db,**args)
        self._refreshSQL = "select * from rdb$collations where rdb$collation_name='%s'"
        self.nameAttribute = "collation_name"

    #--- protected
    def _get_isSystemObj(self):
        return self.system_flag == 1

class FBMException(FBMBaseDBObject):
    def __init__(self,db,**args):
        super(FBMException,self).__init__(db,**args)
        self._refreshSQL = "select * from rdb$exceptions where rdb$exception_name='%s'"
        self._dropSQL = "drop exception %s"
        self.nameAttribute = "exception_name"

    #--- protected
    def _get_isSystemObj(self):
        return self.system_flag is not None

class FBMConstraint(FBMBaseDBObject):
    def __init__(self,db,**args):
        super(FBMConstraint,self).__init__(db,**args)
        self._refreshSQL = "select * from rdb$relation_constraints where rdb$constraint_name='%s'"
        self.nameAttribute = "constraint_name"
    def drop(self):
        # TODO: Reimplement Drop constraints as ALTER TABLE...
        pass

class FBMTrigger(FBMBaseDBObject):
    def __init__(self,db,**args):
        super(FBMTrigger,self).__init__(db,**args)
        self._refreshSQL = "select * from rdb$triggers where rdb$trigger_name='%s'"
        self._dropSQL = "drop trigger %s"
        self.nameAttribute = "trigger_name"

    #--- protected
    def _get_isSystemObj(self):
        return self.system_flag is not None

class FBMIndex(FBMBaseDBObject):
    def __init__(self,db,**args):
        super(FBMIndex,self).__init__(db,**args)
        self._refreshSQL = "select * from rdb$indices where rdb$index_name='%s'"
        self._dropSQL = "drop index %s"
        self.nameAttribute = "index_name"

    #--- protected
    def _get_isSystemObj(self):
        #if self.system_flag is None:
        #    r = [c for c in self.db.constraints if c.index_name is not None and c.index_name.strip() == self.name]
        #else:
        #    r = (1,)
        return self.system_flag or self.db._get_constraintIndices().has_key(self.name)

    #--- Actions
    def recompute(self):
        """Recomputes index statistics."""
        try:
            self.con.execute_immediate("set statistics index %s" % self.name)
            self.con.commit()
            print "Index %s recomputed." % self.name
        except Exception, e:
            print e.__class__, ":", e
    def deactivate(self,verbose=True):
        """Deactivates index."""
        if not self.isSystemObj:
            try:
                self.con.execute_immediate("alter index %s inactive" % self.name)
                self.con.commit()
                if verbose:
                    print "Index %s deactivated." % self.name
            except Exception, e:
                print e.__class__, ":", e
    def activate(self,verbose=True):
        """Activates index."""
        if not self.isSystemObj:
            try:
                self.con.execute_immediate("alter index %s active" % self.name)
                self.con.commit()
                if verbose:
                    print "Index %s activated." % self.name
            except Exception, e:
                print e.__class__, ":", e
    def rebuild(self):
        """Rebuilds index."""
        if not self.isSystemObj:
            self.deactivate(False)
            self.activate(False)
            print "Index %s rebuilt." % self.name


########################################################################
# test units
########################################################################

class TestDatabaseFunctions(unittest.TestCase):
    def testNewFromConnection(self):
        self.con = kdb.connect(dsn="/opt/firebird/examples/employee.fdb",user="SYSDBA",password="masterkey")
        self.db = FBMDatabase(self.con)
        self.db.close()

    def testAttach(self):
        self.db = FBMDatabase()
        self.db.attachDatabase(dsn="/opt/firebird/examples/employee.fdb",user="SYSDBA",password="masterkey")
        self.db.close()

    def testCreateDatabase(self):
        self.db = FBMDatabase()
        self.db.createDatabase("create database '/var/tmp/fbmtest.fdb' user 'SYSDBA' password 'masterkey'")
        self.db.drop()

class TestMetadatFunctions(unittest.TestCase):
    def setUp(self):
        self.db = FBMDatabase()
        self.db.createDatabase("create database '/var/tmp/fbmtest.fdb' user 'SYSDBA' password 'masterkey'")

    def testAttributes(self):
        self.failUnless(len(self.db.domains) >= 0)
        self.failUnless(len(self.db.tables) >= 0)
        self.failUnless(len(self.db.views) >= 0)
        self.failUnless(len(self.db.generators) >= 0)
        self.failUnless(len(self.db.characterSets) >= 0)
        self.failUnless(len(self.db.collations) >= 0)
        self.failUnless(len(self.db.exceptions) >= 0)

    def testDomains(self):
        d = self.db.newDomain("TEST","varchar(20)",characterSet="WIN1250",collate="PXW_CSY",check="VALUE in ('A','N')",
          default="'A'", notNull=True)
        self.failUnlessEqual(d.name, "TEST")
        self.failIf(d.nullable)
        d2 = self.db.newDomain("TESTI","integer")

        print d.attributes
        d.reload()
        self.failIf(d.dropped)
        d2 = self.db.domain("TEST")
        self.failUnlessEqual(d2.name, "TEST")
        d.drop()
        self.failIf(not d.dropped)
        self.failIf(d.attributes)
        self.failIf(d.name)
        d2 = self.db.domain("nonsense")
        self.failIf(d2)

    def testTables(self):
        t = self.db.tables[0]
        t.reload()
        self.failIf(t.dropped)
        t = self.db.table("RDB$DATABASE")
        self.failUnlessEqual(t.name, "RDB$DATABASE")
        self.failUnlessEqual(t.system_flag, 1)
        t2 = self.db.table("nonsense")
        self.failIf(t2)

    def testViews(self):
        v2 = self.db.view("nonsense")
        self.failIf(v2)

    def testGenerator(self):
        g = self.db.newGenerator("TEST")
        self.failUnlessEqual(g.name, "TEST")
        self.failUnlessEqual(g.currentValue, 0)
        self.failUnless(g.attributes)
        g.currentValue = 10
        self.failUnlessEqual(g.currentValue, 10)
        g.reload()
        self.failIf(g.dropped)
        g2 = self.db.generator("TEST")
        self.failUnlessEqual(g.name, g2.name)
        g.drop()
        self.failIf(not g.dropped)
        self.failIf(g.attributes)
        self.failIf(g.name)
        g = self.db.newGenerator("TEST",10)
        self.failUnlessEqual(g.name, "TEST")
        self.failUnlessEqual(g.currentValue, 10)
        g.reload()
        self.failIf(g.dropped)
        g.drop()
        g = self.db.generator("nonsense")
        self.failIf(g)

    def testCharacterSets(self):
        c = self.db.characterSets[0]
        c.reload()
        self.failIf(c.dropped)
        c = self.db.charcterSet("WIN1250")
        self.failUnlessEqual(c.name, "WIN1250")
        self.failUnlessEqual(len(c.collations), 6)
        c2 = self.db.charcterSet("nonsense")
        self.failIf(c2)
        c.drop()
        self.failIf(c.dropped)

    def testCollations(self):
        c = self.db.collations[0]
        c.reload()
        self.failIf(c.dropped)
        c2 = self.db.collation("WIN1250")
        self.failUnlessEqual(c2.name, "WIN1250")
        c2 = self.db.collation("nonsense")
        self.failIf(c2)
        c.drop()
        self.failIf(c.dropped)

    def testExceptions(self):
        e = self.db.newException("TEST","test message")
        self.failUnlessEqual(e.name, "TEST")
        self.failUnlessEqual(e.message, "test message")
        e.reload()
        self.failIf(e.dropped)
        e2 = self.db.exception("TEST")
        self.failUnlessEqual(e2.name, "TEST")
        e.drop()
        self.failIf(not e.dropped)
        self.failIf(e.attributes)
        self.failIf(e.name)
        e2 = self.db.exception("nonsense")
        self.failIf(e2)

    def tearDown(self):
        self.db.drop()

########################################################################
# main program
########################################################################

if __name__ == '__main__':
    print
    print "Firebird Core Classes v%s" % FBCORE_VERSION
    print
    suite = unittest.TestSuite()
    suite.addTest(unittest.makeSuite(TestDatabaseFunctions))
    suite.addTest(unittest.makeSuite(TestMetadatFunctions))
    unittest.TextTestRunner(verbosity=1).run(suite)
